package com.jdbernard.wdiwtlt.cli

import com.jdbernard.wdiwtlt.ConfigWrapper
import com.jdbernard.wdiwtlt.MediaLibrary
import com.jdbernard.wdiwtlt.db.ORM
import com.jdbernard.wdiwtlt.db.models.*
import com.jdbernard.io.NonBlockingInputStreamReader
import com.jdbernard.util.AnsiEscapeCodeSequence as ANSI
import com.zaxxer.hikari.HikariConfig
import com.zaxxer.hikari.HikariDataSource
import org.docopt.Docopt
import jline.console.ConsoleReader
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import java.util.regex.Matcher
import org.slf4j.Logger
import org.slf4j.LoggerFactory

import static com.jdbernard.util.AnsiEscapeCodeSequence.*

public class CommandLineInterface {

    public static final VERSION = "ALPHA"

    public static final def DOC = """\
wdiwtlt v$VERSION

Usage:
    wdiwtlt [options]
    wdiwtlt --version
    wdiwtlt --help

Options:
    -c --config <config-file>

        Config file for wdiwtlt CLI.

    -L --library-root <root-dir>

        The path to a local media library directory. Providing a local library
        causes wdiwtlt to run in local mode.

    -D --database-config <config-file>

        The path to a data source configuration file. wdiwtlt cli uses HikariCP
        for database connection pooling, so the given configuration file is
        expected to be formatted so that it can be fed directly to the
        HikariConfig constructor.
        (see https://github.com/brettwooldridge/HikariCP#configuration-knobs-baby)

Configuration:
"""

    private static Logger logger =
        LoggerFactory.getLogger(CommandLineInterface)

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)

    private Properties cliConfig
    private MediaLibrary library

    /// IO Management
    private InputStream inStream
    private OutputStream outStream

    private ConsoleReader reader
    private Thread consoleReaderThread
    private List<String> consoleReadBuffer =
        Collections.synchronizedList(new ArrayList<String>())
    private synchronized boolean running

    /// Console output data
    private String titleStyle, normalStyle, statusStyle, promptStyle,
        artistStyle, albumStyle, fileStyle, errorStyle, playlistStyle
    private String clearLine = new ANSI().eraseLine(Erase.All).toString()
    private String afterInput =
        new ANSI().eraseLine(Erase.All).scrollUp().cursorUp().toString()
    private String beforeLeader =
        new ANSI().saveCursor().cursorPrevLine(2).toString()
    private String afterLeader =
        new ANSI().restoreCursor().eraseLine(Erase.ToEnd).toString()
    private String eraseLeader =
        new ANSI().eraseLine(Erase.All).cursorPrevLine().eraseLine(Erase.All)
            .cursorPrevLine().eraseLine(Erase.All).toString()

    public final static modelClass = [
        'album': Album, 'artist': Artist, 'bookmark': Bookmark,
        'mediaFile': MediaFile, 'playlist': Playlist, 'tag': Tag ]

    private int displayWidth = 79
    private long msgTimeout
    private ScrollText currentlyPlaying = new ScrollText(
        maxWidth: displayWidth - 15 - VERSION.size(),
        text: "No media currently playing.")
    private ScrollText status = new ScrollText(maxWidth: displayWidth)
    private Date dismissMsgDate = new Date()
    private SimpleDateFormat sdf = new SimpleDateFormat('EEE-HH-SSS')

    /// Current play queue and selection data
    Selection currentSelection = new Selection()
    Playlist playQueue = library.save(new Playlist(
        name: "CLI Queue ${sdf.format(new Date())}"))
    Bookmark playBookmark
    MediaFile curMediaFile

    public static void main(String[] args) {

        def opts = new Docopt(DOC).withVersion("wdiwtlt v$VERSION").parse(args)

        def exitErr = { msg ->
            System.err.println("wdiwtlt: $msg")
            System.exit(1) }

        // Look for a given CLI config file.
        def givenCfg = new Properties()
        File cfgFile
        if (opts["--config"]) cfgFile = new File(opts["--config"])

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File("wdiwtlt.cli.properties")

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File(System.getenv().HOME, ".wdiwtlt.cli.properties")

        if (cfgFile.exists() && cfgFile.isFile()) {
            try { cfgFile.withInputStream { givenCfg.load(it) } }
            catch (Exception e) { givenCfg.clear() } }

        // Just in case, load up the general WDIWTLT default config service
        def wdiwtltDefaultConfig = new ConfigWrapper()

        // Find the library root directory (TODO: eventually this should be
        // enabled only when local mode is chosen).
        File libRoot = new File(
            opts["--library-root"] ?:
            givenCfg[ConfigWrapper.LIBRARY_DIR_KEY] ?:
            wdiwtltDefaultConfig.libraryRootPath ?:
            "no libRoot configured")

        if (libRoot && (!libRoot.exists() || !libRoot.isDirectory()))
            exitErr("Library root does not exist or is not a directory: " +
                libRoot.canonicalPath)

        // Get the library database
        HikariConfig hcfg
        File dbCfgFile = new File(
            opts["--database-config"] ?:
            givenCfg[ConfigWrapper.DB_CONFIG_FILE_KEY] ?:
            "database.properties")

        // Try to load from the given DB config file
        if (dbCfgFile && dbCfgFile.exists() && dbCfgFile.isFile()) {
            Properties props = new Properties()
            try { dbCfgFile.withInputStream { props.load(it) } }
            catch (Exception e) { props.clear() }

            if (props) hcfg = new HikariConfig(props) }

        // Look to see if database connection properties are given in the CLI
        // config file.
        if (givenCfg && givenCfg["database.config.dataSourceClassName"]) {
            Properties props = new Properties()
            props.putAll(givenCfg
                .findAll { it.key.startsWith("database.config.") }
                .collectEntries { [it.key[16..-1], it.value] } )
            if (props) hcfg = new HikariConfig(props) }

        // Fall back to the WDIWTLT general database defaults
        if (!hcfg) hcfg = wdiwtltDefaultConfig.hikariConfig
        if (!hcfg) exitErr("Cannot load database configuation.")

        // Create our database instance
        def orm
        try {
            def hds = new HikariDataSource(hcfg)
            orm = new ORM(hds) }
        catch (Exception e) {
            exitErr("Could not establish a connection to the database:\n\t" +
                e.localizedMessage) }

        // Create our CLI object
        def cliInst = new CommandLineInterface(new MediaLibrary(orm, libRoot),
            System.in, System.out, givenCfg)
        cliInst.repl()
    }

    public CommandLineInterface(MediaLibrary library, InputStream sin,
    OutputStream out, Properties cliConfig) {
        this.cliConfig = cliConfig
        this.library = library
        this.inStream = sin
        this.outStream = out

        this.msgTimeout = cliConfig["message.timeout"] ?: 5000l

        setupTextStyles()

        this.consoleReaderThread = new Thread().start(this.&runReaderThread)
    }

    void setupTextStyles() {
        titleStyle = new ANSI().color(Colors.WHITE, Colors.DEFAULT, false).toString()
        normalStyle = new ANSI().resetText().toString()
        promptStyle = new ANSI().color(Colors.YELLOW, Colors.DEFAULT, true).toString()
        statusStyle = new ANSI().color(Colors.CYAN, Colors.DEFAULT, false).toString()
        playlistStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        artistStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, false).toString()
        albumStyle = new ANSI().color(Colors.BLUE, Colors.DEFAULT, false).toString()
        fileStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        errorStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, true).toString()
    }

    private void runReaderThread() {
        this.reader = new ConsoleReader(this.inStream, this.outStream)
        this.reader.setPrompt(promptStyle + "> " + normalStyle)

        String line
        while(this.running && !Thread.currentThread().isInterrupted()) {
            line = reader.readLine()
            if (line == null || line == "quit" || line == "exit") running = false
            else {
                consoleReadBuffer.add(line)
                outStream.print(afterInput)
                outStream.flush() } } }

    public void repl() {
        this.running = true
        String line = null

        outStream.println ""
        outStream.println ""
        drawLeader()

        while(this.running) {

            if (new Date() > dismissMsgDate) resetStatus()
            if (consoleReadBuffer.size() > 0) {
                line = consoleReadBuffer.remove(0)
                processInput(line.split(/\s/) as LinkedList)
            } else {
                drawLeader()
                Thread.sleep(250)
            }
        }
    }

    private def processInput(LinkedList<String> line) {
        logger.debug("line: $line")
        String command = line.poll()
        logger.debug("command: $command")
        switch(command?.toLowerCase()) {
            case 'scan':        return scanMediaLibrary()
            case 'select':      return processSelect(line, currentSelection)
            case 'list':        return processList(line, currentSelection)
            case 'add':         return processAdd(line)
            case 'enqueue':     return processEnqueue(line)
            case 'tag':         return processTag(line)
            case 'clear':       return processClear(line)
            case 'play':        return line.size() == 0 ?
                                    transport.play() :
                                    processPlay(line)
            case 'pause':       return transport.pause()
            case 'stop':        return transport.stop()
            case 'next':        return processNext(line)
            case 'prev':        return processPrev(line)
            case 'ff':
            case 'fastforward': return processFastForward(line)
            case 'rwd':
            case 'rewind':      return processRewind(line)

            case 'debug':
                outStream.println(
                    "\n\nConfig: \n" +
                    cliConfig.collect { "\t${it.key}: ${it.value}" }
                             .join("\n") +
                    "\n\n\n")
                drawLeader(true)
                return

            case 'q': case 'quit': case ':q': case 'exit': case '\u0004':
                running = false
                consoleReaderThread.interrupt()
                return

            default:
                status.text = errorStyle +
                    "Unrecognized command: '$line'${normalStyle}"
                dismissMsgDate = new Date(new Date().time + msgTimeout)
                drawLeader()
                Thread.sleep(250)
                break
        }
    }

    public MediaLibrary scanMediaLibrary() {
        status.text = "Scanning media library..."
        def counts = library.rescanLibrary()
        status.text = "Scanned ${counts.total} files. " +
            "Added ${counts.new} and ignored ${counts.ignored} files."
        dismissMsgDate = new Date(new Date().time + msgTimeout)
        return library }

    private def processSelect(LinkedList line, def sel = null) {
        String option = line.poll()
        boolean current = option == "current"
        if (!sel) sel = new Selection()
        def items

        if (current) {
            if (!curMediaFile) {
                setErr "No media is currently playing."
                return null }

            option = line.poll()
            switch (option) {
                case 'album':
                    sel.album = ensureExactlyOne(
                        library.getAlbumsWhere({ mediaFileId: curMediaFile.id}))
                    return sel
                case 'artist':
                    sel.artist = ensureExactlyOne(
                        library.getArtistsWhere({ mediaFileId: curMediaFile.id}))
                    return sel
                case 'playlist':
                    sel.playlist = playQueue; return sel
                case 'file':
                    sel.mediaFile = curMediaFile; return sel
                case 'tags':
                    sel.tags = library.getTagsWhere({
                        mediaFileId: curMediaFile.id})
                    return sel
                default:
                    setErr("Unrecognized option to ${promptStyle}select " +
                        "current${errorStyle}.")
                    return null
            }
        }

        switch (option) {
            case 'file': option = 'mediaFile'
            case 'album': case 'artist': case 'playlist':
                sel[option] = ensureExactlyOne(
                    library.getByIdOrName(modelClass[option], line.join(' ')))
                return sel
            case 'tags':
                sel.tags = line.collect { library.getByIdOrName(Tag, it) }
                               .findAll().flatten()
                return sel

            default:
                setErr("Unrecognized option to ${promptStyle}select${errorStyle}")
                return null
        }
    }

    private def processAdd(LinkedList line) {

        List<String> parts = line.join(' ').split(' to playlist ').collect { it.trim() }

        if (parts.size() != 2) {
            printLongMessage(
                "Invalid options to the ${promptStyle}add${normalStyle}" +
                " command. Use ${promptStyle}help add${normalStyle} to see a " +
                "list of valid options.")
            return }

        Selection selectionToAdd
        Playlist targetPlaylist

        if (parts[0] == "selection") { selectionToAdd = currentSelection }
        else selectionToAdd = processSelect(parts[0].split(' ') as LinkedList)

        targetPlaylist = ensureExactlyOne(
            library.getByIdOrName(Playlist, parts[1]))

        return library.addToPlaylist(targetPlaylist.id,
            selection.selectedFiles.collect { it.id }) }

    private def processEnqueue(LinkedList line) {
        Selection selection = processSelect(line)
        return library.addToPlaylist(playQueue.id,
            selection.selectedFiles.collect { it.id }) }

    private def processTag(LinkedList line) {
        List<String> parts = line.join(' ').split(' as ').collect { it.trim() }

        if (parts.size() != 2) {
            printLongMessage(
                "Invalid options to the ${promptStyle}tag${normalStyle}" +
                " command. Use ${promptStyle}help tag${normalStyle} to see a " +
                "list of valid options.")
            return }


        Selection selection = processSelect(parts[0].split(' ') as LinkedList)
        library.tagMediaFiles(selection.selectedFiles.collect { it.id },
            parts[1].split(' ') as List) }

    private def processClear(LinkedList line) {
        def option = line.poll()
        switch(option) {
            case 'queue': return library.removeAllFromPlaylist(playQueue.id)
            case 'selected playlist':
                if (!currentSelection.playlist) {
                    printLongMessage("No playlist currently selected.")
                    return null }
                return library.removeAllFromPlaylist(selected.playlist.id)
            case 'playlist':
                def criteria = line.poll()
                if (!criteria) {
                    printLongMessage("Missing playlist id or name. Use " +
                        "${promptStyle}help clear${normalStyle} to see a " +
                        "list of valid options.")
                    return null }
                def playlist = library.getByIdOrName(Playlist, criteria)
                if (!playlist) {
                    printLongMessage("No playlist matches '${criteria}'.")
                    return null }
                return library.removeAllFromPlaylist(playlist.id)
            case 'selection':
                currentSelection = new Selection()
                break
            default:
                printLongMessage("Unrecognized option to the ${promptStyle}" +
                    "clear${normalStyle} command. Use ${promptStyle}help clear" +
                    "${normalStyle} to see a list of valid options.")
                return null
        }
    }

    private def processList(LinkedList options, def selection) {
        logger.debug("Listing things. Options: $options")
        if (!selection) selection = new Selection()
        def option = options.poll()
        boolean all = option == 'all'
        if (all) option = options.poll()

        logger.debug("Option: $option")

        def list

        switch(option) {
            case 'albums':
                if (all) list = library.getAlbums()
                else if (selection.album) list = [selection.album]
                else list = library.getAlbumsWhere(
                    playlistId: selection.playlist?.id,
                    artistId: selection.artist?.id,
                    mediaFileId: selection.mediaFile?.id)

                String albumMatch = options?.join(" ")?.trim()
                if (albumMatch) list = list.findAll { it.name =~ albumMatch }
                printLongMessage(makeList(selection, Album, list, all,
                    { "${it.id}: ${it}" }))
                break

            case 'artists':
                if (all) list = library.getArtists()
                else if (selection.artist) list = [selection.artist]
                else list = library.getArtistsWhere(
                    playlistId: selection.playlist?.id,
                    albumId: selection.album?.id,
                    mediaFileId: selection.mediaFile?.id)

                String artistMatch = options?.join(" ")?.trim()
                if (artistMatch) list = list.findAll { it.name =~ artistMatch }
                printLongMessage(makeList(selection, Artist, list, all,
                    { "${it.id}: ${it.name}" }))
                break

            case 'files':
            case 'selection':
                if (all) list = library.getMediaFiles()
                else list = selection.selectedFiles

                if (selection.album) list = list.sort { it.trackNumber }
                String mediaFileMatch = options?.join(" ")?.trim()
                if (mediaFileMatch) list = list.findAll {
                    it.name =~ mediaFileMatch }
                printLongMessage(makeList(selection, MediaFile, list, all,
                    { "${it.id}: ${it.trackNumber} - ${it.name}" }))
                break

            case 'bookmarks':
                if (all) list = library.getBookmarks()
                else list = library.getBookmarksWhere(
                    playlistId: selection?.playlist?.id)

                String bookmarkMatch = options?.join(" ")?.trim()
                if (boolmarkMatch)
                     list = list.findAll { it.name =~ bookmarkMatch }
                printLongMessage(makeList(selection, Bookmark, list, all,
                    { "${it.id}: ${it.name} ${it.userCreated ? '' : ' (auto)'}" }))
                break

            case 'playlists':
                if (all) list = library.getPlaylists()
                else list = library.getPlaylistsWhere(
                    artistId: selection?.artist?.id,
                    albumId: selection?.album?.id,
                    mediaFileId: selection?.mediaFile?.id)

                String playlistMatch = options?.join(" ")?.trim()
                if (playlistMatch)
                     list = list.findAll { it.name =~ playlistMatch }
                printLongMessage(makeList(selection, Playlist, list, all,
                    { "${it.id}: ${it.name} ${it.userCreated ? '' : ' (auto)'}" }))
                break

            case 'tags':
                if (all) list = library.getTags()
                else list = library.getTagsWhere(
                    playlistId: selection?.playlist?.id,
                    artistId: selection?.artist?.id,
                    albumId: selection?.album?.id)

                String tagMatch = options?.join(" ")?.trim()
                if (tagMatch) list = list.findAll { it.name =~ tagMatch }
                printLongMessage(makeList(selection, Tag, list, all,
                    { "${it.id}: ${it.name}" }))
                break

            default:
                printLongMessage("Unrecognized option to the ${promptStyle}" +
                    "list${normalStyle} command. Use ${promptStyle}help list" +
                    "${normalStyle} to see a list of valid options.")
                return null
        }

        resetStatus()
        return list
    }

    public def ensureExactlyOne(def matches) {
        if (!matches) {
            setErr("Nothing matches.");
            return null }

        String englishName = toEnglish(matches[0].class.simpleName)
        if (matches.size() > 1) {
            setErr("Multiple  ${englishName}s match: " +
                matches.collect { "${it.id}: ${it.name}" }.join(', '))
            return null }

        return matches[0] }

    public def getExactlyOne(Class modelClass, def criteria) {
        return ensureExactlyOne(library.getByIdOrName(modelClass, criteria)) }

    private void drawLeader(afterOutput = false) {

        String leader = beforeLeader + getLeader() +
            (afterOutput ?
                (promptStyle + "> " + normalStyle) :
                afterLeader)

        outStream.print(leader)
        outStream.flush() }

    private void printLongMessage(String msg) {
        String result = new StringBuilder()
            .append(eraseLeader)
            .append(msg)
            .append("\n\n")
            .toString()

        outStream.println result
        drawLeader(true) }

    private String getLeader() {
        StringBuilder leader = new StringBuilder()
            .append(clearLine)
            .append(titleStyle)
            .append("WDIWTLT - v")
            .append(VERSION)
            .append(normalStyle)
            .append(" -- ")
            .append(statusStyle)
            .append(currentlyPlaying)
            .append("\n")
            .append(clearLine)
            .append(normalStyle)
            .append(status)
            .append("\n")

        return leader.toString() }

    private String setErr(String errMsg) {
        status.text = errorStyle + errMsg
        dismissMsgDate = new Date(new Date().time + msgTimeout) }

    private String makeList(Selection selection, Class modelClass,
    def items, boolean listAll = false, Closure toString = null) {

        def result = new StringBuilder()
            .append("--------------------\n${modelClass.simpleName}s")

        if (!listAll && (selection.playlist || selection.artist ||
                         selection.mediaFile))
            result.append("\n(for selection: ")
                  .append(selection.toString())
                  .append(normalStyle)
                  .append(")")

        result.append(":\n\n")
        result.append(items.collect( toString ?: { it.toString() }).join("\n"))
            .append("\n")

        return result.toString() }

    private String resetStatus() {
        String s = currentSelection.toString()

        if (s.size() == 0) status.text = "No current media selections."
        else status.text = s

        return status.text }

    private static String uncapitalize(String s) {
        if (s == null) return null;
        if (s.length() < 2) return s.toLowerCase();
        return s[0].toLowerCase() + s[1..-1] }

    private static String toEnglish(String modelName) {
        return UPPERCASE_PATTERN.matcher(modelName).
            replaceAll(/$1 $2/).toLowerCase() }

    public class Selection {
        Album album
        Artist artist
        MediaFile mediaFile
        Playlist playlist
        List<Tag> tags

        public def unselect(Class modelClass) {
            String key = uncapitalize(modelClass.simpleName)
            def value = this[key]
            this[key] = null
            return value }

        public Selection select(Map s) {
            ['artist', 'album', 'playlist', 'file', 'tags'].each {
                this[it] = s[it] }
            return this
        }

        public def select(def value) {
            if (value) this[uncapitalize(match.class.simpleName)] = value
            return value }

        public List<MediaFile> getSelectedFiles() {
            return library.getMediaFilesWhere(
                playlistId: playlist?.id,
                artistId: artist?.id,
                albumId: album?.id,
                tags: tags) }

        public String toString() {
            StringBuilder s = new StringBuilder()

            if (playlist) s.append(playlistStyle)
                .append(playlist)
                .append(normalStyle)
                .append(": ")

            if (artist) s.append(artistStyle)
                .append(artist)
                .append(normalStyle)
                .append(" / ")

            if (album) s.append(albumStyle)
                .append(album)
                .append(normalStyle)
                .append(" / ")

            if (mediaFile) s.append(fileStyle)
                .append(mediaFile)
                .append(normalStyle)

            return s.toString()
        }


    }
}
